# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
from .....schema_classes import GenericAspectClass
from .....schema_classes import GenericPayloadClass
from .....schema_classes import MetadataChangeEventClass
from .....schema_classes import MetadataChangeLogClass
from .....schema_classes import MetadataChangeProposalClass
from .....schema_classes import PlatformEventClass
from .....schema_classes import PlatformEventHeaderClass
from .....schema_classes import SystemMetadataClass


GenericAspect = GenericAspectClass
GenericPayload = GenericPayloadClass
MetadataChangeEvent = MetadataChangeEventClass
MetadataChangeLog = MetadataChangeLogClass
MetadataChangeProposal = MetadataChangeProposalClass
PlatformEvent = PlatformEventClass
PlatformEventHeader = PlatformEventHeaderClass
SystemMetadata = SystemMetadataClass

# fmt: on
